/*****************************************************************************
 * Copyright (c) 2009, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "kite_object.h"
#include "kite_vm.h"

/*****************************************************************************
 * Call operator method on object programmatically
 ****************************************************************************/
int kite_vm_call_operator(kite_thread_t *thd, struct kite_object_t *this, 
                          int op, struct kite_object_t *arg, int err)
{
    kite_object_t *cur = this, *exc;
    kite_symtab_t *st;
    
    if (cur->type < OBJ_NUMBER_OF_BASE_TYPES)
    {
        switch(cur->type)
        {
            case OBJ_INTEGER:
            {
                cur = kite_dereference_and_load(thd, "System.integer");
                break;
            }
            case OBJ_FLOAT:
            {
                cur = kite_dereference_and_load(thd, "System.float");
                break;
            }
            case OBJ_BOOLEAN:
            {
                cur = kite_dereference_and_load(thd, "System.boolean");
                break;
            }
            default:
            {
                assert(0);
            }
        }
    }
    
    do {
        KITE_FIND_ANY_IN_SYMTAB(st, cur->object_data.properties, OP_TO_METHOD(op));
        if (st) {
            kite_vm_call_object(thd, this, st->value, arg);
            return 1;
        }
        
        cur = cur->object_data.inherit_from;
    } while(cur);
    
    /* Operator not defined, raise exception */
    if (err) {
        exc = kite_new_exception(thd, "System.exceptions.OperatorNotDefined", 
                                 OP_TO_STRING(op));
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
    }
    return 0;
}

/*****************************************************************************
 * Call constructor on object programmatically
 ****************************************************************************/
void kite_vm_call_constructor(kite_thread_t *thd, struct kite_object_t *this, 
                              struct kite_object_t *args) 
{
    kite_symtab_t *symt;
    kite_object_t *curr = this->object_data.inherit_from;
    
    /* 
     * Find __construct__ (with any arguments) in the object first. If we
     * can't find any, just ignore this call. Otherwise, call constructor
     * and have the system throw an exception if the arguments don't match
     * up.
     *
     * Note [Ticket 5]: we're only looking in the current object, not the
     * parents. Any constructors in the current object are responsible for
     * calling parent constructors as needed (by using 
     * classname.__construct__).
     *
     * Note2: We also want to spit out an error if we're trying to call a
     * constructor that takes arguments that does not exist.
     */    
    KITE_FIND_ANY_IN_SYMTAB(symt, curr->object_data.properties, 
        "__construct__");
    if (symt != NULL || kite_list_count(thd, args) > 0) {
        kite_vm_call_method(thd, this, "__construct__", args, TRUE);
        return;
    }
}

/*****************************************************************************
 * Call method on object programmatically
 ****************************************************************************/
void kite_vm_call_method(kite_thread_t *thd, kite_object_t *obj, 
                         char *name, struct kite_object_t *args, int err)
{
    kite_symtab_t *found = NULL;
    kite_object_t *curr = obj, *exc;
    char buf[1024];
    int numargs = kite_list_count(thd, args);
    
    if (curr->type < OBJ_NUMBER_OF_BASE_TYPES)
    {
        switch(curr->type)
        {
            case OBJ_INTEGER:
            {
                curr = kite_dereference_and_load(thd, "System.integer");
                break;
            }
            case OBJ_FLOAT:
            {
                curr = kite_dereference_and_load(thd, "System.float");
                break;
            }
            case OBJ_BOOLEAN:
            {
                curr = kite_dereference_and_load(thd, "System.boolean");
                break;
            }
            default:
            {
                assert(0);
            }
        }
    }
    
    do {
        KITE_FIND_METHOD(found, curr, name, numargs);
        if (found && found->value->type == OBJ_METHOD) {
            kite_vm_call_object(thd, obj, found->value, args);
            return;
        } else if (found) {
            sprintf(buf, "%s is not callable.", name);
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                                     buf);
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
        
        curr = curr->object_data.inherit_from;
    } while(curr);
    
    /* Call function operator if method's not found */
    if (strcmp("__construct__", name) != 0 &&
        strcmp("__destruct__", name) != 0) {
        curr = kite_new_list(thd);
        curr->builtin_data.listvalue.car = kite_new_string(thd, name);
        curr->builtin_data.listvalue.cdr = kite_new_list(thd);
        curr->builtin_data.listvalue.cdr->builtin_data.listvalue.car =
            kite_reference_object(args);
            
        if (kite_vm_call_operator(thd, obj, OP_METHOD_CALL, curr, 0)) {
            kite_dereference_object(curr);
            return;
        }
        
        kite_dereference_object(curr);
    }
    
    /* Method not found, raise exception */
    if (err) {
        if (strcmp("__construct__", name) == 0) {
            sprintf(buf, "Could not find constructor that takes %d argument(s)",
                    numargs);
        } else {
            sprintf(buf, "Could not find method %s that takes %d argument(s)",
                    name, numargs);
        }
        exc = kite_new_exception(thd, "System.exceptions.ObjectNotFound", 
                                 buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
}